# -*- coding: utf-8 -*-

import signal
import setproctitle

from .connection import Connection
from ..share.log import logger
from .request import Request
from ..share import constants


class Worker(object):

    type = constants.PROC_TYPE_WORKER

    request_class = Request
    connection_class = Connection
    got_first_request = False

    group_id = None

    # 是否有效(父进程中代表程序有效，子进程中代表worker是否有效)
    enable = True

    def __init__(self, app, group_id):
        """
        构造函数
        :return:
        """
        self.app = app
        self.group_id = group_id

    def run(self):
        setproctitle.setproctitle(self.app.make_proc_name(
            '%s:%s' % (self.type, self.group_id)
        ))

        self._handle_proc_signals()
        self._on_worker_run()

        try:
            address = self.app.config['WORKER_IPC_ADDRESS_TPL'] % self.group_id
            conn = self.connection_class(self, address, self.app.config['WORKER_CONN_TIMEOUT'])
            conn.run()
        except KeyboardInterrupt:
            pass
        except:
            logger.error('exc occur.', exc_info=True)

    def _on_worker_run(self):
        self.app.events.create_worker(self)
        for bp in self.app.blueprints:
            bp.events.create_app_worker(self)

    def _handle_proc_signals(self):
        def exit_handler(signum, frame):
            # 防止重复处理KeyboardInterrupt，导致抛出异常
            if self.enable:
                self.enable = False
                raise KeyboardInterrupt

        def safe_stop_handler(signum, frame):
            self.enable = False

        # 强制结束，抛出异常终止程序进行
        signal.signal(signal.SIGINT, exit_handler)
        signal.signal(signal.SIGQUIT, exit_handler)
        # 安全停止
        signal.signal(signal.SIGTERM, safe_stop_handler)
        signal.signal(signal.SIGHUP, safe_stop_handler)




# -*- coding: utf-8 -*-

import os
import copy
import json
import sys
import subprocess
import time
import signal
import setproctitle
from netkit.box import Box
from netkit.contrib.tcp_client import TcpClient
import thread

from ..share.log import logger
from ..share.utils import safe_call
from ..share.thread_timer import ThreadTimer
from ..share import constants


class Master(object):
    """
    master相关
    """

    type = constants.PROC_TYPE_MASTER

    app = None

    enable = True

    # 等待重启所有workers的timer
    restart_workers_timer = None

    # proxy进程列表
    proxy_process = None

    # worker进程列表
    worker_processes = None

    def __init__(self, app):
        """
        构造函数
        :return:
        """
        self.app = app
        self.restart_workers_timer = ThreadTimer()

    def run(self):
        setproctitle.setproctitle(self.app.make_proc_name(self.type))

        self._handle_proc_signals()

        self._spawn_proxy()
        self._spawn_workers()

        thread.start_new(self._connect_to_proxy, ())

        self._monitor_child_processes()

    def _connect_to_proxy(self):
        """
        连接到proxy，因为有些命令要发过来
        :return:
        """
        client = TcpClient(Box, address=self.app.config['MASTER_IPC_ADDRESS'])

        while True:
            try:
                client.connect()
            except KeyboardInterrupt:
                break
            except:
                # 只要连接失败
                logger.error('connect fail. address: %s', self.app.config['MASTER_IPC_ADDRESS'])
                time.sleep(1)
                continue

            # 读取的数据
            box = client.read()
            if not box:
                logger.info('connection closed.')
                continue

            logger.info('box: %s', box)

            safe_call(self._handle_proxy_data, box)

    def _handle_proxy_data(self, box):
        """
        处理从proxy过来的box
        :param box:
        :return:
        """

        if box.cmd == constants.CMD_ADMIN_CHANGE_GROUP:
            jdata = json.loads(box.body)
            group_id = jdata['payload']['group_id']
            count = jdata['payload']['count']

            # 不能设置成个奇怪的值就麻烦了
            assert isinstance(count, int), 'data: %s' % box.body

            if group_id not in self.app.config['GROUP_CONFIG']:
                self.app.config['GROUP_CONFIG'][group_id] = dict(
                    count=count
                )

            else:
                self.app.config['GROUP_CONFIG'][group_id]['count'] = count

            self._restart_workers()

        elif box.cmd == constants.CMD_ADMIN_RELOAD_WORKERS:
            self._reload_workers()
        elif box.cmd == constants.CMD_ADMIN_RESTART_WORKERS:
            self._restart_workers()
        elif box.cmd == constants.CMD_ADMIN_STOP:
            self._safe_stop()

    def _start_child_process(self, proc_env):
        worker_env = copy.deepcopy(os.environ)
        worker_env.update({
            self.app.config['CHILD_PROCESS_ENV_KEY']: json.dumps(proc_env)
        })

        args = [sys.executable] + sys.argv
        inner_p = subprocess.Popen(args, env=worker_env)
        inner_p.proc_env = proc_env
        return inner_p

    def _spawn_proxy(self):
        proc_env = dict(
            type=constants.PROC_TYPE_PROXY
        )
        self.proxy_process = self._start_child_process(proc_env)

    def _spawn_workers(self):
        self.worker_processes = []

        for group_id, group_info in self.app.config['GROUP_CONFIG'].items():
            proc_env = dict(
                type=constants.PROC_TYPE_WORKER,
                group_id=group_id,
            )

            # 进程个数
            for it in xrange(0, group_info['count']):
                p = self._start_child_process(proc_env)
                self.worker_processes.append(p)

    def _monitor_child_processes(self):
        while 1:
            if self.proxy_process and self.proxy_process.poll() is not None:
                proc_env = self.proxy_process.proc_env
                if self.enable:
                    self.proxy_process = self._start_child_process(proc_env)

            for idx, p in enumerate(self.worker_processes):
                if p and p.poll() is not None:
                    # 说明退出了
                    proc_env = p.proc_env
                    self.worker_processes[idx] = None

                    if self.enable and not self.restart_workers_timer.is_set():
                        # 如果还要继续服务
                        p = self._start_child_process(proc_env)
                        self.worker_processes[idx] = p

            if not filter(lambda x: x, self.worker_processes):
                # 没活着的了worker了

                if self.restart_workers_timer.is_set():
                    # 如果是在等待重启，就直接重启了
                    self.restart_workers_timer.clear()
                    self._spawn_workers()
                    continue
                else:
                    break

            # 时间短点，退出的快一些
            time.sleep(0.1)

    def _restart_workers(self):
        """
        安全停止所有workers
        :return:
        """

        for p in self.worker_processes:
            if p:
                p.send_signal(signal.SIGTERM)

        def final_kill_workers():
            """
            如果到时间还没停止，那就只能强制kill了
            """
            for p in self.worker_processes:
                if p:
                    p.send_signal(signal.SIGKILL)

        self.restart_workers_timer.set(self.app.config['STOP_TIMEOUT'], final_kill_workers)

    def _reload_workers(self):
        """
        reload是热更新，停一个，起一个
        :return:
        """
        for p in self.worker_processes:
            if p:
                p.send_signal(signal.SIGHUP)

    def _safe_stop(self):
        """
        安全停止所有子进程，并最终退出
        如果退出失败，要最终kill -9
        :return:
        """
        self.enable = False

        for p in self.worker_processes + [self.proxy_process]:
            if p:
                p.send_signal(signal.SIGTERM)

        if self.app.config['STOP_TIMEOUT'] is not None:
            signal.alarm(self.app.config['STOP_TIMEOUT'])

    def _handle_proc_signals(self):
        def exit_handler(signum, frame):
            self.enable = False

            # 如果是终端直接CTRL-C，子进程自然会在父进程之后收到INT信号，不需要再写代码发送
            # 如果直接kill -INT $parent_pid，子进程不会自动收到INT
            # 所以这里可能会导致重复发送的问题，重复发送会导致一些子进程异常，所以在子进程内部有做重复处理判断。
            for p in self.worker_processes + [self.proxy_process]:
                if p:
                    p.send_signal(signum)

            # https://docs.python.org/2/library/signal.html#signal.alarm
            if self.app.config['STOP_TIMEOUT'] is not None:
                signal.alarm(self.app.config['STOP_TIMEOUT'])

        def final_kill_handler(signum, frame):
            if not self.enable:
                # 只有满足了not enable，才发送term命令
                for p in self.worker_processes + [self.proxy_process]:
                    if p:
                        p.send_signal(signal.SIGKILL)

        def safe_stop_handler(signum, frame):
            """
            等所有子进程结束，父进程也退出
            """
            self._safe_stop()

        def safe_reload_handler(signum, frame):
            """
            让所有子进程重新加载
            """
            self._reload_workers()

        # INT, QUIT为强制结束
        signal.signal(signal.SIGINT, exit_handler)
        signal.signal(signal.SIGQUIT, exit_handler)
        # TERM为安全结束
        signal.signal(signal.SIGTERM, safe_stop_handler)
        # HUP为热更新
        signal.signal(signal.SIGHUP, safe_reload_handler)
        # 最终判决，KILL掉子进程
        signal.signal(signal.SIGALRM, final_kill_handler)



# -*- coding: utf-8 -*-

"""
从haven拿过来的
"""

import threading
from utils import safe_call


class ThreadTimer(object):

    timer = None

    def set(self, interval, callback, repeat=False, force=True):
        """
        添加timer
        """
        if self.timer:
            if force:
                # 如果已经存在，那么先要把现在的清空
                self.clear()
            else:
                # 已经存在的话，就返回了
                return

        def callback_wrapper():

            # 必须要确定，这次调用就是这个timer引起的
            if self.timer == timer:
                self.timer = None
                result = safe_call(callback)
                if repeat and not self.timer:
                    # 之所以还要判断timer，是因为callback中可能设置了新的回调
                    self.set(interval, callback, repeat, True)
                return result

        self.timer = timer = threading.Timer(interval, callback_wrapper)
        # 跟着主线程结束。默认值是和创建的线程一致
        timer.daemon = True
        timer.start()

    def is_set(self):
        """
        是否被设置了timer
        """
        return self.timer is not None

    def clear(self):
        """
        直接把现在的清空
        """
        if not self.timer:
            return

        try:
            self.timer.cancel()
        except:
            pass
        self.timer = None



# -*- coding: utf-8 -*-

from collections import defaultdict

from ..share import constants
from group_queue import GroupQueue


class TaskDispatcher(object):
    """
    任务管理
    主要包括: 消息来了之后的分发
    """

    # 之所以不用WeakSet的原因是，经过测试worker断掉之后，对象不会立即被删除，极有有可能会被用到。
    # 繁忙worker列表
    busy_workers_dict = None
    # 空闲
    idle_workers_dict = None
    # 消息队列
    group_queue = None

    def __init__(self):
        self.busy_workers_dict = defaultdict(set)
        self.idle_workers_dict = defaultdict(set)
        self.group_queue = GroupQueue()

    def remove_worker(self, worker):
        """
        删除worker，一般是worker断掉了
        :param worker:
        :return:
        """
        if worker in self.busy_workers_dict[worker.group_id]:
            self.busy_workers_dict[worker.group_id].remove(worker)
            return

        if worker in self.idle_workers_dict[worker.group_id]:
            self.idle_workers_dict[worker.group_id].remove(worker)
            return

    def add_task(self, group_id, item):
        """
        添加任务
        当新消息来得时候，应该先检查有没有空闲的worker，如果没有的话，才放入消息队列
        :return:
        """
        idle_workers = self.idle_workers_dict[group_id]
        if not idle_workers:
            self.group_queue.put(group_id, item)
            return

        # 弹出一个可用的worker
        worker = idle_workers.pop()
        # 变成处理中
        worker.status = constants.WORKER_STATUS_BUSY
        # 放到队列中去
        self.busy_workers_dict[group_id].add(worker)

        # 让worker去处理任务吧
        worker.assign_task(item)

    def alloc_task(self, worker):
        """
        尝试获取新任务
        :return: 获取的新任务
        """
        task = self.group_queue.get(worker.group_id)
        dst_status = constants.WORKER_STATUS_BUSY if task else constants.WORKER_STATUS_IDLE

        if worker.status != dst_status:
            # 说明状态有变化，需要调整队列
            worker.status = dst_status
            # 同步状态
            self._sync_worker_status(worker)

        return task

    def _sync_worker_status(self, worker):
        """
        内部 同步worker的状态：空闲/繁忙
        此时worker的status，已经自己改过了
        :param worker:
        :return:
        """

        if worker.status == constants.WORKER_STATUS_BUSY:
            src_workers_dict = self.idle_workers_dict
            dst_workers_dict = self.busy_workers_dict
        else:
            src_workers_dict = self.busy_workers_dict
            dst_workers_dict = self.idle_workers_dict

        if worker in src_workers_dict[worker.group_id]:
            # 因为有可能worker的状态是None的话，是不在任何队列里面的，所以先判断一下
            src_workers_dict[worker.group_id].remove(worker)

        dst_workers_dict[worker.group_id].add(worker)




# -*- coding: utf-8 -*-

NAME = 'burst'

# 系统返回码
# 命令字不合法
RET_INVALID_CMD = -10000
# 系统内部异常
RET_INTERNAL = -10001
# admin用户验证失败
RET_ADIMN_AUTH_FAIL = -20000
# master连接未连接
RET_MASTER_NOT_CONNECTED = -21000

# 内部使用的命令字
# 分配任务. 如果body里面带数据，说明是要写回；如果没有数据，说明只是要分配task
CMD_WORKER_TASK_ASSIGN = 100
# 任务完成
CMD_WORKER_TASK_DONE = 200

# 管理员命令
# 获取运行状态统计
CMD_ADMIN_SERVER_STAT = 20000

# 修改group配置，比如worker数量。workers会自动restart
CMD_ADMIN_CHANGE_GROUP = 21000
# 重新加载workers
CMD_ADMIN_RELOAD_WORKERS = 21001
# 重启workers，与RELOAD不同，是先全部停止，再启动
CMD_ADMIN_RESTART_WORKERS = 21002
# 停止整个server
CMD_ADMIN_STOP = 21003


# worker的状态
WORKER_STATUS_IDLE = 1
WORKER_STATUS_BUSY = 2

# 进程类型
PROC_TYPE_MASTER = 'master'
PROC_TYPE_PROXY = 'proxy'
PROC_TYPE_WORKER = 'worker'


# 默认配置
DEFAULT_CONFIG = {
    # 监听IP
    'HOST': '127.0.0.1',
    # 监听端口
    'PORT': 9900,

    # 进程名
    'NAME': NAME,

    # 是否调试模式
    'DEBUG': False,

    # box class
    'BOX_CLASS': 'netkit.box.Box',

    # master class
    'MASTER_CLASS': 'burst.master.Master',

    # proxy class
    'PROXY_CLASS': 'burst.proxy.Proxy',

    # worker class
    'WORKER_CLASS': 'burst.worker.Worker',

    # 分组进程配置(group_id务必为数字):
    #    {
    #        $group_id: {
    #            count: 10,
    #        }
    #    }
    'GROUP_CONFIG': {
        1: {
            'count': 1,
        }
    },
    # 通过box路由group_id:
    #    def group_router(box):
    #        return group_id
    'GROUP_ROUTER': lambda box: 1,

    # 停止子进程超时(秒). 使用 TERM 进行停止时，如果超时未停止会发送KILL信号
    'STOP_TIMEOUT': None,

    # master<->worker之间通信的address
    'MASTER_IPC_ADDRESS': NAME + '_ipc/master.sock',

    # proxy<->worker之间通信的address模板
    'WORKER_IPC_ADDRESS_TPL': NAME + '_ipc/%s.sock',

    # proxy的backlog
    'PROXY_BACKLOG': 256,

    # worker<->proxy网络连接超时(秒), 包括 connect once，read once，write once
    'WORKER_CONN_TIMEOUT': 3,
    # 处理task超时(秒). 超过后worker会自杀. None 代表永不超时
    'WORK_TIMEOUT': None,
    # worker重连等待时间
    'WORKER_TRY_CONNECT_INTERVAL': 1,

    # 子进程标识进程类型的环境变量
    'CHILD_PROCESS_ENV_KEY': 'BURST_ENV',

    # 管理员，可以连接proxy获取数据
    # 管理员访问地址: NAME + '_ipc/admin.sock' or ('127.0.0.1', 9910)
    'ADMIN_ADDRESS': NAME + '_ipc/admin.sock',
    'ADMIN_USERNAME': None,
    'ADMIN_PASSWORD': None,

    # 统计相关
    # 作业时间统计标准
    'TASKS_TIME_BENCHMARK': (10, 50, 100, 500, 1000, 5000),
}
